print("import: 'astroid'")
import astroid

print("import: 'astroid.brain'")
import astroid.brain

print("import: 'astroid.interpreter'")
import astroid.interpreter

print("import: 'astroid.nodes'")
import astroid.nodes

print("import: 'astroid.modutils'")
import astroid.modutils

